/*
 *  gen_cert.c
 *  generate a new X.509 formatted cerificate.
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <openssl/x509.h>
#include <openssl/evp.h>
#include <openssl/rsa.h>
#include "gen_cert.h"



/*
 *  Get RSA-2048 Key
 */



EVP_PKEY* genkey() {
  EVP_PKEY* key;
  key = EVP_PKEY_new();

  EVP_PKEY_assign_RSA(key, RSA_generate_key(2048, RSA_F4, NULL, NULL)); //RSA_F4 is 65537

  return key;
}



/*
 *  Fill Out General Certificate Attributes
 */



X509_NAME* make_name(entity* info) {
  X509_NAME* name;
  
  X509_NAME_add_entry_by_txt(name, "C",  MBSTRING_ASC, (unsigned char *)info->country, -1, -1, 0);
  X509_NAME_add_entry_by_txt(name, "O",  MBSTRING_ASC, (unsigned char *)info->org, -1, -1, 0);
  X509_NAME_add_entry_by_txt(name, "CN",  MBSTRING_ASC, (unsigned char *)info->common_name, -1, -1, 0);

  if (info->org_unit) {
    X509_NAME_add_entry_by_txt(name, "OU",  MBSTRING_ASC, (unsigned char *)info->org_unit, -1, -1, 0);
  }

  if (info->state) {
    X509_NAME_add_entry_by_txt(name, "ST",  MBSTRING_ASC, (unsigned char *)info->state, -1, -1, 0);
  }

  if (info->locality) {
    X509_NAME_add_entry_by_txt(name, "L",  MBSTRING_ASC, (unsigned char *)info->locality, -1, -1, 0);
  }
  
  return name;
}

void fill_cert(X509* cert, EVP_PKEY* key) {
  ASN1_INTEGER_set(X509_get_serialNumber(cert), 1);     //Set serial number to 1
  X509_gmtime_adj(X509_get_notBefore(cert), 0);         //Set not before to now
  X509_gmtime_adj(X509_get_notAfter(cert), 31536000L);  //Set not after to a year from now

  X509_set_pubkey(cert, key);
}



/*
 *  Write PK And Cert
 */ 



void write_pk_and_cert(pk_and_cert* pk_ct, X509* cert, EVP_PKEY* key) {
  FILE* fp_pk;
  fp_pk = fmemopen(pk_ct->private_key, 512, "wb");

  FILE* fp_cert;
  fp_cert = fmemopen(pk_ct->certificate, 512, "wb");
  
  PEM_write_PrivateKey(fp_pk, key, NULL, NULL, NULL, NULL, NULL);
  PEM_write_X509(fp_cert, cert);

  fclose(fp_pk);
  fclose(fp_cert);
}



/*
 *  Generate Self-Signed Certificate
 *  http://stackoverflow.com/a/15082282/4151046 
 */


pk_and_cert* gen_ss_cert(entity* info) {
  pk_and_cert* pk_ct;
  EVP_PKEY* key = genkey();

  X509* cert = X509_new();
  fill_cert(cert, key);

  X509_NAME* name = make_name(info);
  
  x509_set_issuer_name(cert, name);
  x509_set_subject_name(cert, name);
  
  x509_sign(cert, key, EVP_sha512());

  write_pk_and_cert(pk_ct, cert, key);

  return pk_ct;
}



/*
 *  Generate CA-Signed Certificate
 *  http://stackoverflow.com/a/15082282/4151046 
 */



pk_and_cert* gen_ca_cert(entity* subject, X509* ca_cert, EVP_PKEY* ca_key) {
  pk_and_cert* pk_ct;
  EVP_PKEY* key = genkey();

  X509* cert = X509_new();
  fill_cert(cert, key);

  X509_NAME* subject = make_name(subject);

  x509_set_issuer_name(cert, X509_get_subject_name(ca_cert));
  x509_set_subject_name(cert, subject);

  x509_sign(cert, ca_key, EVP_sha512());

  write_pk_and_cert(pk_ct, cert, key);

  return pk_ct;
}
