/*
 * dhe_rsa.c
 * key exchange using DHE_RSA, where RSA is RSA-2048
 * samuel judson | cryptography and network protocols | marlboro college 
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <openssl/sha.h>
#include <gmp.h>
#include "dhe_rsa.h"
#include "rsa/rsa.h"
#include "dhe/dhe.h"
#include "keygen/keygen.h"

#ifndef RSA_SPECS
#include "rsa/rsa_specs.h"
#define RSA_SPECS 1
#endif

#ifndef DHE_SPECS
#include "dhe/dhe_specs.h"
#define DHE_SPECS 1
#endif

#define T 1
#define F 0

#define HEX_CHARS_PER_BYTE 2



/*
 *  Get Or Generate Key
 */



void read_rsa_keyfile(char* path, keys* key_ptr, int pri) {
  FILE *fp;
  fp = fopen(path, "rb");

  if (pri) {
    gmp_fscanf(fp, "%ZX.%ZX", key_ptr->m_pri_exp, key_ptr->m_n);
  } else {
    gmp_fscanf(fp, "%ZX.%ZX", key_ptr->o_pub_exp, key_ptr->o_n);
  }
}

void generate_rsa_key(keys* key_ptr) {
  FILE* fp_pub;
  fp_pub = fopen("rsa_key.pub", "wb");

  FILE* fp_pri;
  fp_pri = fopen("rsa_key.pri", "wb");

  mpz_inits(key_ptr->m_n, key_ptr->m_pri_exp, key_ptr->m_pub_exp, NULL);

  rsa_keygen(key_ptr->m_n, key_ptr->m_pri_exp, key_ptr->m_pub_exp);
  
  gmp_fprintf(fp_pub, "%ZX.%ZX", key_ptr->m_pub_exp, key_ptr->m_n);
  gmp_fprintf(fp_pri, "%ZX.%ZX", key_ptr->m_pri_exp, key_ptr->m_n);

  fprintf(stdout, "\nRSA key pair generated.\n\n");
}



/*
 *  Hash Payload (SHA-512)
 */



void hash_payload(mpz_t hashed, mpz_t payload) {
  int i;

  char payload_str[RSA_MODULUS_LENGTH * HEX_CHARS_PER_BYTE];  //not going to be hashing anything larger
  unsigned char hash[SHA512_DIGEST_LENGTH];
  unsigned char hex_hash[(SHA512_DIGEST_LENGTH * HEX_CHARS_PER_BYTE) + 1];

  gmp_sprintf(payload_str, "%ZX", payload);
  SHA512(payload_str, sizeof(payload_str) - 1, hash);

  for (i = 0; i < SHA512_DIGEST_LENGTH; i++) {
    sprintf(&hex_hash[i * HEX_CHARS_PER_BYTE], "%02X", hash[i]);
  }

  gmp_sscanf(hex_hash, "%ZX", hashed);
}



/*
 *  Generate And Sign Party Secret
 */



void gen_party_secret(mpz_t party_secret, mpz_t hashed_party_secret, mpz_t signed_secret, keys* key_ptr, mpz_t dhe_rand_sec, mpz_t dhe_generator, mpz_t dhe_prime) {
  dhe_compute_party_secret(party_secret, dhe_rand_sec, dhe_generator, dhe_prime);
  hash_payload(hashed_party_secret, party_secret);
  rsa_sign_payload(signed_secret, party_secret, key_ptr->m_pri_exp, key_ptr->m_n);
}



/*
 *  Verify And Combine Other Party Secret
 */



void gen_shared_secret(mpz_t shared_secret, keys* key_ptr, mpz_t other_party_secret, mpz_t other_party_secret_hash, mpz_t dhe_rand_sec, mpz_t dhe_prime) {
  mpz_t verified_other_party_secret, verified_other_party_secret_hash;
  mpz_inits(verified_other_party_secret, verified_other_party_secret_hash, NULL);

  rsa_verify_signature(verified_other_party_secret, other_party_secret, key_ptr->o_pub_exp, key_ptr->o_n);
  hash_payload(verified_other_party_secret_hash, verified_other_party_secret);

  if (!mpz_cmp(verified_other_party_secret_hash, other_party_secret_hash)) {
    dhe_compute_shared_secret(shared_secret, dhe_rand_sec, verified_other_party_secret, dhe_prime);
  } else {
    perror("Error: Invalid signature");
    exit(-1);
  }
  
  mpz_clears(verified_other_party_secret, verified_other_party_secret_hash, NULL);
}
