/*  
 *  serpent_lib/test.c
 *  testing different components of serpent_lib
 *  samuel judson | cryptography and network protocols | marlboro college
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "encrypt.h"
#include "decrypt.h"

#ifndef SPECS
#include "cipher_specs.h"
#define SPECS 1
#endif



/*
 *  PAD TEST VECTORS - RANDOMLY GENERATED BLOCK
 */



char* PAD_TEST_BLOCK = "997E82920D6F568394A3";
char* PAD_TEST_BLOCK_PADDED = "997E82920D6F568394A3FFFFFFFFFFFF";

//bytes, will be padded to 16
#define PAD_TEST_BLOCK_LENGTH      10 

/*
 *  CIPHER TEST VECTORS - VECTOR 1, SET 4 FROM SERPENT TESTS (http://www.cs.technion.ac.il/~biham/Reports/Serpent/Serpent-256-128.verified.test-vectors)
 */

char* CIPHER_TEST_KEY = "2BD6459F82C5B300952C49104881FF482BD6459F82C5B300952C49104881FF48";
char* CIPHER_TEST_PLAINTEXT = "EA024714AD5C4D84EA024714AD5C4D84";
char* CIPHER_TEST_CIPHERTEXT = "3E507730776B93FDEA661235E1DD99F0";



/*
 *  Pad And Unpad Test
 */



void pad_test() {
  int i, j, k, last_byte_loc;
  unsigned char chunk_str[BYTE_CHUNK_SIZE];
  unsigned char block_str[BLOCK_LENGTH * BYTE_CHUNK_SIZE];

  uint32_t block[BLOCK_LENGTH];
  uint32_t compare_block[BLOCK_LENGTH];

  // Convert testing string to block.
  for (i = 0; i < BLOCK_LENGTH; i++) {
    memcpy(chunk_str, PAD_TEST_BLOCK + (i * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE), BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE);
    sscanf(chunk_str, "%08X", &block[i]);
  }

  pad(block, (BLOCK_LENGTH * BYTE_CHUNK_SIZE) - PAD_TEST_BLOCK_LENGTH);

  fprintf(stdout, "\n===== PADDING TESTS =====\n\n");

  for (k = 0; k < BLOCK_LENGTH; k++) {
    sprintf(block_str + (k * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE), "%08X", block[k]);
  } 

  if (!strcmp(block_str, PAD_TEST_BLOCK_PADDED)) {
    fprintf(stdout, "PAD TEST: SUCCESS.\n\n");
  } else {
    fprintf(stdout, "PAD TEST: FAILURE.\n\n");
  }

  fprintf(stdout, "Expected:   %s\nGot:        %s\n\n", PAD_TEST_BLOCK_PADDED, block_str);
 
  /* END PAD TEST -------- BEGIN UNPAD TEST */

  // Convert testing string to block.
  for (j = 0; j < BLOCK_LENGTH; j++) {
    memcpy(chunk_str, PAD_TEST_BLOCK_PADDED + (j * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE), BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE);
    sscanf(chunk_str, "%08X", &compare_block[j]);
  }

  last_byte_loc = unpad(compare_block);

  if (last_byte_loc == PAD_TEST_BLOCK_LENGTH) {
    fprintf(stdout, "UNPAD TEST: SUCCESS.\n\n");
  } else {
    fprintf(stdout, "UNPAD TEST: FAILURE.\n\n");
  }

  fprintf(stdout, "Expected Last Byte Index:   %d\nGot:                        %d\n\n", PAD_TEST_BLOCK_LENGTH, last_byte_loc);
}



/*
 *  Encrypt Test
 */



void encrypt_test() {
  FILE* temp_out;
  FILE* enc_res_in;

  unsigned char chunk_str[BYTE_CHUNK_SIZE];
  unsigned char key_str[KEY_LENGTH * BYTE_CHUNK_SIZE];
  uint32_t key[KEY_LENGTH];

  char* enc_res;
  int i;

  // Convert testing key to key block.
  for (i = 0; i < KEY_LENGTH; i++) {
    memcpy(chunk_str, CIPHER_TEST_KEY + (i * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE), BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE);
    sscanf(chunk_str, "%08X", &key[i]);
  }

  enc_res = serpent_encrypt(key, CIPHER_TEST_PLAINTEXT, 1);

  fprintf(stdout, "===== ENCRYPTION TEST =====\n\n");

  if (!strcmp(enc_res, CIPHER_TEST_CIPHERTEXT)) {
    fprintf(stdout, "ENCRYPT TEST: SUCCESS.\n\n");
  } else {
    fprintf(stdout, "ENCRYPT TEST: FAILURE.\n\n");
  }
  
  fprintf(stdout, "Expected: %s\nGot:      %s\n\n", CIPHER_TEST_CIPHERTEXT, enc_res);
}



/*
 *  Decrypt Test
 */



void decrypt_test() {
  FILE* temp_out;
  FILE* dec_res_in;

  unsigned char chunk_str[BYTE_CHUNK_SIZE];
  unsigned char key_str[KEY_LENGTH * BYTE_CHUNK_SIZE];
  uint32_t key[KEY_LENGTH];

  char* dec_res;
  int i;

  // Convert testing key to key block.
  for (i = 0; i < KEY_LENGTH; i++) {
    memcpy(chunk_str, CIPHER_TEST_KEY + (i * BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE), BYTE_CHUNK_SIZE * HEX_CHARS_PER_BYTE);
    sscanf(chunk_str, "%08X", &key[i]);
  }

  dec_res = serpent_decrypt(key, CIPHER_TEST_CIPHERTEXT, 1);

  fprintf(stdout, "===== DECRYPTION TEST =====\n\n");

  if (!strcmp(dec_res, CIPHER_TEST_PLAINTEXT)) {
    fprintf(stdout, "DECRYPT TEST: SUCCESS.\n\n");
  } else {
    fprintf(stdout, "DECRYPT TEST: FAILURE.\n\n");
  }
  
  fprintf(stdout, "Expected: %s\nGot:      %s\n\n", CIPHER_TEST_PLAINTEXT, dec_res);
}


int main() {
  pad_test();
  encrypt_test();
  decrypt_test();

  return 0;
}
