/* 
 * x509.h
 * writes and parses X.509 certificates - code taken, with additions and alterations, from Davies' Implementing SSL/TLS Using Cryptography and PKI.
 * sam judson | cryptography and network protocols | marlboro college
 */


#ifndef X509_H
#define X509_H

#include <time.h>
#include <gmp.h>

#include "rsa.h"
#include "dsa.h"

typedef enum {
  rsa,
  dh
} algorithmIdentifier;

typedef enum {
  sha512withRSAEncryption,
} signatureAlgorithmIdentifier;

/**
 * A name (or "distinguishedName") is a list of attribute-value pairs.
 * Instead of keeping track of all of them, just keep track of
 * the most interesting ones.
 */
typedef struct {
  char *idAtCountryName;
  char *idAtStateOrProvinceName;
  char *idAtLocalityName;
  char *idAtOrganizationName;
  char *idAtOrganizationalUnitName;
  char *idAtCommonName;
} name;

typedef struct {
  // TODO deal with the "utcTime" or "GeneralizedTime" choice.
  time_t notBefore;
  time_t notAfter;
} validity_period;

typedef mpz_t uniqueIdentifier;

typedef struct {
  algorithmIdentifier algorithm;
  mpz_t rsa_public_key_exp;
  mpz_t rsa_public_key_mod;
} public_key_info;

typedef mpz_t objectIdentifier;
typedef struct {
  int version;
  mpz_t serialNumber; // This can be much longer than a 4-byte long allows
  signatureAlgorithmIdentifier signature;
  name issuer;
  validity_period validity;
  name subject;
  public_key_info subjectPublicKeyInfo;
  uniqueIdentifier issuerUniqueId;
  uniqueIdentifier subjectUniqueId;
  int certificate_authority; // 1 if this is a CA, 0 if not
} x509_certificate;

typedef struct {
  x509_certificate tbsCertificate;
  unsigned int *hash; // hash code of tbsCertificate
  int hash_len;
  signatureAlgorithmIdentifier algorithm;
  mpz_t rsa_signature_value;
} signed_x509_certificate;

void init_x509_certificate(signed_x509_certificate *certificate);
int parse_x509_certificate(const unsigned char *buffer,  const unsigned int certificate_length, signed_x509_certificate *parsed_certificate);
void free_x509_certificate(signed_x509_certificate *certificate);

#endif
