/*
 *  keygen/keygen.c
 *  generates keys for use with DHE and RSA-2048
 *  samuel judson | cryptography and network protocols | marlboro college
 */


#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <gmp.h>

#ifndef RSA_SPECS
#include "../rsa/rsa_specs.h"
#define RSA_SPECS 1
#endif

#ifndef DHE_SPECS
#include "../dhe/dhe_specs.h"
#define dhe_SPECS 1
#endif

#define T 1
#define F 0 

//bytes in uint32_t
#define RAND_GEN_SIZE 4

#define HEX_CHAR_PER_BYTE 2



/*
 *  Generate Random Numbers W/ Optional Primality Checking
 */



void get_rand(mpz_t rand_num, int num_length, FILE* urand) {
  uint32_t rand_int;
  int i;  
  int urand_calls = num_length / RAND_GEN_SIZE;

  unsigned char urand_call_response[RAND_GEN_SIZE * HEX_CHAR_PER_BYTE];

  unsigned char* num_hex_str;
  num_hex_str = calloc(HEX_CHAR_PER_BYTE, num_length);
 
  for (i = 0; i < urand_calls; i++) {
    fread(&rand_int, sizeof(uint32_t), 1, urand);
    
    if (i == 0) {
      rand_int = rand_int | (1 << 31); // Make first bit a 1 so that candidate is of correct size.
    }

    sprintf(urand_call_response, "%08X", rand_int);
    strcat(num_hex_str, urand_call_response);
  }

  if (mpz_init_set_str(rand_num, num_hex_str, 16)) {
    perror("Error: Could not generate random number.\n");
    exit(-1);
  }

  free(num_hex_str);
}

void get_prime(mpz_t candidate, int prime_length) {
  int is_prime = F;

  fprintf(stdout, "Generating prime... ");
  fflush(stdout);

  FILE* urand;
  urand = fopen("/dev/urandom", "rb");
  while (!is_prime) {
    get_rand(candidate, prime_length, urand);

    if (mpz_probab_prime_p(candidate, 64) > 0) { //uses Miller-Rabin and trial division.
     is_prime = T;
    }
  }
  
  fclose(urand);
  fprintf(stdout, "Done.\n");
}



/*
 *  RSA Key Generation
 */



int small_difference_check(mpz_t p, mpz_t q, mpz_t n) {
  int passes;

  mpz_t diff, min_size;
  mpz_inits(diff, min_size, NULL);

  mpz_root(min_size, n, 4);
  mpz_mul_ui(min_size, min_size, (unsigned long)2);

  if (mpz_cmp(p, q) > 0) {
    mpz_sub(diff, p, q);
  } else {
    mpz_sub(diff, q, p);
  }

  if (mpz_cmp(diff, min_size) > 0) {
    passes = T;
  } else {
    passes = F;
  }

  mpz_clears(diff, min_size, NULL);
  return passes;
}

int weiner_low_exp_check(mpz_t p, mpz_t q, mpz_t n, mpz_t pri_exp) {
  int passes = T;

  mpz_t trip_exp, quad_root_n, doub_prime;
  mpz_inits(trip_exp, quad_root_n, doub_prime, NULL);

  mpz_mul_ui(trip_exp, pri_exp, (unsigned long)3);
  mpz_root(quad_root_n, n, 4);
  
  if (mpz_cmp(quad_root_n, trip_exp) > 0) {

    if (mpz_cmp(p, q) > 0) {
      mpz_mul_ui(doub_prime, q, (unsigned long)2);

      if (mpz_cmp(doub_prime, p) > 0) {
	passes = F;
      }

    } else {
      mpz_mul_ui(doub_prime, p, (unsigned long)2);

      if (mpz_cmp(doub_prime, q) > 0) {
	passes = F;
      }
    }
  }
 
  mpz_clears(trip_exp, quad_root_n, doub_prime, NULL);
  return passes;
}

//https://en.wikipedia.org/wiki/RSA_(cryptosystem)#Faulty_key_generation
int component_weakness_checks(mpz_t p, mpz_t q, mpz_t n, mpz_t pri_exp) {
  int passes;

  if (!(passes = small_difference_check(p, q, n))) {
    return F;
  } 

  if (!(passes = weiner_low_exp_check(p, q, n, pri_exp))) {
    return F;
  }

  return T;
}

void rsa_keygen(mpz_t n, mpz_t pri_exp, mpz_t pub_exp) {
  int passes = F;

  mpz_t p, q, totient_n, p_min_one, q_min_one;
  mpz_inits(p, q, totient_n, p_min_one, q_min_one, NULL);

  get_prime(p, RSA_PRIME_LENGTH);

  while (!passes) {
    get_prime(q, RSA_PRIME_LENGTH);
    mpz_mul(n, p, q);

    mpz_init_set_ui(pub_exp, (unsigned long)RSA_PUB_EXP); 

    // compute private key exponent  
    mpz_sub_ui(p_min_one, p, (unsigned long)1);
    mpz_sub_ui(q_min_one, q, (unsigned long)1);
    mpz_mul(totient_n, p_min_one, q_min_one);
    mpz_invert(pri_exp, pub_exp, totient_n);

    passes = component_weakness_checks(p, q, n, pri_exp);
  }
  
  mpz_clears(p, q, totient_n, p_min_one, q_min_one, NULL);
}



/*
 *  DHE Key Generation
 */



void dhe_keygen(mpz_t rand_sec) {
  FILE* urand;
  urand = fopen("/dev/urandom", "rb");
  
  get_rand(rand_sec, DHE_KEY_LENGTH, urand);
  fclose(urand);
}









